/*
 * Copyright 2003-2005 Broadcom Corporation. All rights reserved.
 * 
 * Redistribution and use of this Software in source and binary forms,
 * with or without modification, are permitted provided that the following
 * conditions are met:
 * 
 * 	1. Redistributions of source code must retain the above copyright
 * 	   notice, this list of conditions and the following disclaimer. 
 * 
 * 	2. Redistributions in binary form must reproduce the above copyright
 * 	   notice, this list of conditions and the following disclaimer in
 * 	   the documentation and/or other materials provided with the
 * 	   distribution.
 * 
 * 	3. This Software may be used only in connection with Broadcom
 * 	   storage integrated circuits.
 * 
 * THIS SOFTWARE IS PROVIDED BY BROADCOM CORPORATION AS IS AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL BROADCOM CORPORATION OR ITS LICENSORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be used, modified and/or distributed under the terms of
 * the GNU General Public License, version 2, as published by the Free
 * Software Foundation (the GPL), in which case the provisions of the GPL
 * apply INSTEAD OF those given above. A copy of the GPL is available at
 * http://www.gnu.org/copyleft/gpl.html, or by writing to the 
 * 	Free Software Foundation, Inc.
 * 	59 Temple Place - Suite 330
 * 	Boston, MA 02111-1307, USA.
 */ 

#include "version.h"
#include "build_number.h"
#include "stamp.h"

#include <linux/version.h>
#include <linux/config.h>
#include <linux/module.h>

#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/hdreg.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/proc_fs.h>

#include <linux/smp.h>

#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/io.h>


static int bccfg_major;

#define MAJOR_NR           bccfg_major
#define DEVICE_NR(device)  MINOR(device)
#define DEVICE_NAME        "bccfg"


#include <linux/blk.h>
#include <linux/blkpg.h>
#include "scsi.h"
#include "hosts.h"

#include <scsi/scsi_ioctl.h>
#include <scsi/sg.h>


#define BCCFG_VERSION              RC_VERSION_STR
#define BCCFG_BUILD_DATE	       __DATE__
#define BCCFG_BUILD_TIME	       __TIME__
#define BCCFG_NAME	               "bccfg"
#define BCCFG_MAX_DEVS             8
#define BCCFG_MAX_RETRIES          5
#define BCCFG_TIMEOUT              (30 * HZ)


MODULE_AUTHOR("Broadcom, Inc");
MODULE_DESCRIPTION("Configuration Driver for Broadcom raid controllers");
MODULE_LICENSE("Dual BSD/GPL");

static int debug = 0;
#ifdef module_param_named
module_param_named(debug, debug, int, 0444);
#else
MODULE_PARM (debug, "i");
#endif
MODULE_PARM_DESC (debug, "debug print level");


#define DBG_PRINTK(...) if (debug) printk(__VA_ARGS__);

/* support for /proc/scsi/bccfg/version
 */
static struct proc_dir_entry *proc_bccfg = NULL;
static struct proc_dir_entry *proc_bccfg_version = NULL;
static int bccfg_proc_read(char *page, char **start, off_t off, int count, int *peof, void *data);
static int bccfg_proc_write(struct file *file, const char *buf, unsigned long count, void *data);

/*
 * Private configuration device data. 
 */
struct bccfg_dev {
        struct scsi_device *device;
	unsigned capacity;              /* size in blocks */
};


static struct bccfg_dev bccfg_devs[BCCFG_MAX_DEVS];


/*
 *   SCSI Device Operations 
 */

static int bccfg_detect(struct scsi_device *);
static int bccfg_init(void);
static void bccfg_finish(void);
static int bccfg_attach(struct scsi_device *);
static void bccfg_detach(struct scsi_device *);
static int bccfg_init_command(Scsi_Cmnd *);
static void bccfg_rw_intr(Scsi_Cmnd * SCpnt);


static struct Scsi_Device_Template bccfg_template = {
	name:BCCFG_NAME,
	tag:DEVICE_NAME,
	scsi_type:TYPE_PROCESSOR,
	dev_max:BCCFG_MAX_DEVS,
	blk:1,
	detect:bccfg_detect,
	init:bccfg_init,
	finish:bccfg_finish,
	attach:bccfg_attach,
	detach:bccfg_detach,
	init_command:bccfg_init_command,
};


/*
 *    bccfg_log_result() - Prints the decoded result field and
 *    the sense buffer to the system log.
 */
static void
bccfg_log_result(Scsi_Request *sreq)
{
        int result = sreq->sr_result;

       	printk("bccfg status = %x, message = %02x, host = %d, driver = %02x\n",
	       status_byte(result),
	       msg_byte(result),
	       host_byte(result),
	       driver_byte(result));

	if ( (CHECK_CONDITION & status_byte(result)) ||
       	    (DRIVER_SENSE & driver_byte(result)) ) {
	        print_req_sense("bccfg", sreq);
	}
}


/*
 * Block Device Operations
 */

/* Open function for the config device. */
static int
bccfg_open(struct inode *inode, struct file *filp)
{
	int minor;
	struct scsi_device *sdev;
	minor = DEVICE_NR(inode->i_rdev);

	DBG_PRINTK("Openning %s%d\n", DEVICE_NAME, minor);

	if (minor >= bccfg_template.dev_max)
		return -ENXIO;	/* No such device */

	sdev = bccfg_devs[minor].device;
 
	if (!sdev)
		return -ENXIO;

	/* If the device is in error recovery, wait until it is done. */
	if (!scsi_block_when_processing_errors(sdev))
		return -ENXIO;

	/* Prevent module unloading while the device is in use */
	if (sdev->host->hostt->module)
		__MOD_INC_USE_COUNT(sdev->host->hostt->module);
	if (bccfg_template.module)
		__MOD_INC_USE_COUNT(bccfg_template.module);

	sdev->access_count++;

	return 0;
}

/* Close function for the config device */ 
static int
bccfg_release(struct inode *inode, struct file *file)
{
	int minor;
	struct scsi_device *sdev;

	minor = DEVICE_NR(inode->i_rdev);

	DBG_PRINTK("Closing %s%d\n", DEVICE_NAME, minor);

	if (minor > bccfg_template.dev_max) 
		return -ENODEV;
	
	sdev = bccfg_devs[minor].device;

	if (!sdev)
		return -ENODEV;

	sdev->access_count--;

	if (sdev->host->hostt->module)
		__MOD_DEC_USE_COUNT(sdev->host->hostt->module);
	if (bccfg_template.module)
		__MOD_DEC_USE_COUNT(bccfg_template.module);

	return 0;
}

static int
bccfg_sg_io(struct scsi_device *sdev, void *args)
{

	struct scsi_request *sreq;
	sg_io_hdr_t  hdr;
	unsigned char direction;
	unsigned char *buf = NULL;
	unsigned char cmd[MAX_COMMAND_SIZE];
	int result = 0;

	if (copy_from_user(&hdr, args, sizeof(sg_io_hdr_t)))
		return -EFAULT;

	if (hdr.interface_id != 'S')
		return -EINVAL;

	if (hdr.cmd_len > MAX_COMMAND_SIZE)
		return -EINVAL;

	if (hdr.iovec_count)
		return -EOPNOTSUPP;

	if (copy_from_user(cmd, hdr.cmdp, hdr.cmd_len))
		return -EFAULT;

	if (debug)
		print_command(cmd);

	if (hdr.dxfer_len) {

		buf = kmalloc(hdr.dxfer_len, GFP_KERNEL);

		if (!buf)
			return -ENOMEM;

		switch (hdr.dxfer_direction) {

		case SG_DXFER_TO_DEV:
			direction = SCSI_DATA_WRITE;
			if (copy_from_user(buf, hdr.dxferp, hdr.dxfer_len)) {
				kfree(buf);
				return -EFAULT;
			}
			break;

		case SG_DXFER_FROM_DEV:
			direction = SCSI_DATA_READ;
			if (!access_ok(VERIFY_WRITE, hdr.dxferp, hdr.dxfer_len)) {
				kfree(buf);
				return -EFAULT;
			}
			break;

		default:
			return -EINVAL;
		}
	}
	else {
		direction = SCSI_DATA_NONE;
	}

	sreq = scsi_allocate_request(sdev);

	if (!sreq) {
		printk(KERN_WARNING "bccfg_sg_io: request allocation failed\n");
		if (direction != SCSI_DATA_NONE)
			kfree(buf);
		return -ENOMEM;
	}
     
	sreq->sr_data_direction = direction;
    
	scsi_wait_req(sreq, (void *)cmd, (void *)buf, hdr.dxfer_len,
		      BCCFG_TIMEOUT, BCCFG_MAX_RETRIES);

	if (sreq->sr_result) {
	        DBG_PRINTK("bccfg%d Error doing scsi request\n", sdev->id);
		bccfg_log_result(sreq);
	}

	hdr.status = 0xff & sreq->sr_result;
	hdr.masked_status = status_byte(sreq->sr_result);
	hdr.msg_status = msg_byte(sreq->sr_result);
	hdr.host_status = host_byte(sreq->sr_result);
	hdr.driver_status = driver_byte(sreq->sr_result);

	/* Copy the sense buffer back to the user. */
	if ((hdr.mx_sb_len > 0) && hdr.sbp) {
		if ((CHECK_CONDITION & hdr.masked_status) ||
		    (DRIVER_SENSE & hdr.driver_status)) {
			int sb_len = 8, len;
			sb_len = (hdr.mx_sb_len > sb_len) ? sb_len : hdr.mx_sb_len;
			/* Additional sense length field */
			len = 8 + (int)sreq->sr_sense_buffer[7];
			len = (len > sb_len) ? sb_len : len;
			copy_to_user(hdr.sbp, sreq->sr_sense_buffer, len);
			hdr.sb_len_wr = len;
		}
	}

	copy_to_user(args, &hdr, sizeof(sg_io_hdr_t));

	if (direction == SCSI_DATA_READ)
		copy_to_user(hdr.dxferp, sreq->sr_buffer, hdr.dxfer_len);

	scsi_release_request(sreq);
     
	if (direction != SCSI_DATA_NONE)
		kfree(buf);

	return 0;
}

static int
bccfg_ioctl(struct inode *inode,
	    struct file *file,
	    unsigned int cmd,
	    unsigned long arg)
{
	struct scsi_device *sdev;
	struct bccfg_dev *bccfg;

	bccfg = &bccfg_devs[DEVICE_NR(inode->i_rdev)];

	if (!bccfg->device)
		return -ENODEV;

	if(!scsi_block_when_processing_errors(bccfg->device))
		return -ENODEV;

	switch (cmd) {

	  /* Fake out the geometery. Can't call into the driver to get it because
	   * bios_param() call in 2.4 kernels need a Scsi_Disk pointer as the first
	   * argument.
           */
	case HDIO_GETGEO:
	{
		struct hd_geometry *loc = (struct hd_geometry *) arg;
		int diskinfo[4];

		if(!loc)
			return -EINVAL;
	
		/* default to most commonly used values */
		diskinfo[0] = 0x40;
		diskinfo[1] = 0x20;
		diskinfo[2] = bccfg->capacity >> 11;

		if (put_user(diskinfo[0], &loc->heads) ||
		    put_user(diskinfo[1], &loc->sectors) ||
		    put_user(diskinfo[2], &loc->cylinders) ||
		    put_user(0, &loc->start))
			return -EFAULT;

		return 0;
	}

	case HDIO_GETGEO_BIG:
	{
		struct hd_big_geometry *loc = (struct hd_big_geometry *)arg;
		int diskinfo[4];
         
		if(!loc)
			return -EINVAL;
         
		/* default to most commonly used values */
		diskinfo[0] = 0x40;
		diskinfo[1] = 0x20;
		diskinfo[2] = bccfg->capacity >> 11;
         
		if (put_user(diskinfo[0], &loc->heads) ||
		    put_user(diskinfo[1], &loc->sectors) ||
		    put_user(diskinfo[2], (unsigned int *) &loc->cylinders) ||
		    put_user(0, &loc->start))
			return -EFAULT;
		return 0;
	}

	case BLKGETSIZE:
	case BLKGETSIZE64:
	case BLKROSET:
	case BLKROGET:
	case BLKRASET:
	case BLKRAGET:
	case BLKFLSBUF:
	case BLKSSZGET:
	case BLKPG:
	case BLKELVGET:
	case BLKELVSET:
	case BLKBSZGET:
	case BLKBSZSET:
		return blk_ioctl(inode->i_rdev, cmd, arg);
         
	case SG_IO:
		return bccfg_sg_io(bccfg->device, (void *)arg);

	default:
		return scsi_ioctl(bccfg->device, cmd, (void *)arg);
	}
}



static struct block_device_operations bccfg_bops =
{
	owner:			THIS_MODULE,
	open:			bccfg_open,
	release:		bccfg_release,
	ioctl:			bccfg_ioctl,
};



/*
 * Find the bccfg device associated with a SCSI Device
*/
static struct bccfg_dev *
bccfg_get_dev(struct scsi_device *sdev)
{
	int i;

	for (i = 0; i < bccfg_template.dev_max; i++)
		if (bccfg_devs[i].device == sdev)
			return &bccfg_devs[i];

	return NULL;

}


/*
 *   bccfg_detect() - Returns one of the scsi device is a 
 *   processor device and vendor string is RAIDCORE
 */
static int
bccfg_detect(struct scsi_device *sdev)
{
	if (sdev->type == TYPE_PROCESSOR &&
	    memcmp(sdev->vendor, "RAIDCORE", 8) == 0) {
		bccfg_template.dev_noticed++;
		return 1;
	}
	return 0;
}


/*
 * The bccfg_init() function looks at all SCSI drives present, determines
 * their size, and reads partition table entries for them.
 */
static int
bccfg_init()
{
	int i;

	if (bccfg_template.dev_noticed == 0)
		return 0;

	DBG_PRINTK("bccfg_init: Found %d configuration devices\n",
		   bccfg_template.dev_noticed); 

	return 0;

}


/* 
 *     bccfg_find_queue() - Callback routine called by the block layer
 *     to determine the queue associated with the bccfg major device.
 */
static request_queue_t *
bccfg_find_queue(kdev_t dev)
{

	if (!bccfg_devs[DEVICE_NR(dev)].device)
		return NULL;

	return &bccfg_devs[DEVICE_NR(dev)].device->request_queue;
}


/* 
 *    bccfg_test_unit_ready() - Send a Test Unit Ready to the config device.
 *    Returns zero if the device is OK or -1 otherwise.
 */
static int
bccfg_test_unit_ready(struct scsi_device *sdev, int retries)
{
        Scsi_Request *sreq;
	unsigned char cmd[10];
	int result = -1;

	
    sreq = scsi_allocate_request(sdev);

	if (!sreq) {
                printk(KERN_WARNING "bccfg_test_unit_ready: Request allocation failure.\n");
	        return -1;
	}

	do {
	    cmd[0] = TEST_UNIT_READY;
		cmd[1] = (sdev->scsi_level <= SCSI_2) ? ((sdev->lun << 5) & 0xe0) : 0;
		memset((void *) &cmd[2], 0, 8);
		sreq->sr_cmd_len = 0;
		sreq->sr_sense_buffer[0] = 0;
		sreq->sr_sense_buffer[2] = 0;
		sreq->sr_data_direction = SCSI_DATA_NONE;

		scsi_wait_req(sreq, (void *)cmd, NULL, 0, BCCFG_TIMEOUT, BCCFG_MAX_RETRIES);

		result = sreq->sr_result;
	} while (retries-- && result);

	scsi_release_request(sreq);

	if (result) {
	        printk(KERN_WARNING "bccfg%d Error doing Unit Test Ready\n", sdev->id);
		bccfg_log_result(sreq);
		return -1;
	}
	return 0;
}


/*
 *    bccfg_read_capacity() - Read the devices cpacity. This function returns
 *    size of the device in 512 byte sectors. If it can't be retrieved return
 *    a reasonable value.
 */
static int
bccfg_read_capacity(struct scsi_device *sdev, int retries)
{
        Scsi_Request *sreq;
	unsigned char cmd[10];
	unsigned char *buf;
	int result = -1;
	int sector_size = 512;
	int capacity =  0x1fffff;


	sreq = scsi_allocate_request(sdev);
	if (!sreq) {
		printk(KERN_WARNING "bccfc_read_capacity: Request allocation failure.\n");
		return capacity;
	}


	buf = (unsigned char *)scsi_malloc(512);
	if (!buf) {
		printk(KERN_WARNING "bccfg_read_capacity: Memory allocation failure.\n");
		scsi_release_request(sreq);
		return capacity;
	}

	do {
		cmd[0] = READ_CAPACITY;
		cmd[1] = (sdev->scsi_level <= SCSI_2) ? ((sdev->lun << 5) & 0xe0) : 0;
		memset((void *) &cmd[2], 0, 8);
		memset((void *) buf, 0, 8);
		sreq->sr_cmd_len = 0;
		sreq->sr_sense_buffer[0] = 0;
		sreq->sr_sense_buffer[2] = 0;
		sreq->sr_data_direction = SCSI_DATA_READ;

		scsi_wait_req(sreq, (void *)cmd, (void *)buf,
			      8, BCCFG_TIMEOUT, BCCFG_MAX_RETRIES);

		result = sreq->sr_result;
	} while (retries-- && result);
	

	if (result) {
	        printk(KERN_WARNING "bccfg%d Error reading capacity\n", sdev->id);
		bccfg_log_result(sreq);
	}
	else {

	        capacity = 1 + ((buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3]);
		sector_size = (buf[4] << 24) | (buf[5] << 16) | (buf[6] << 8) | buf[7];

		/* Rescale capacity to 512-byte sectors */
		switch (sector_size) {

		case 0:
		        sector_size = 512;
			printk(KERN_WARNING "%s%d: sector size 0 reported, assuming 512.\n",
			       DEVICE_NAME, sdev->id);
			break;
		case 256:
		        capacity >>= 1;
			break;
		case 512:
		        break;
		case 1024:
		        capacity <<= 1;
		        break;
		case 2048:
		        capacity <<= 2;
		        break;
		case 4096:
		        capacity <<= 3;
		        break;
		default:
		      printk(KERN_WARNING "%s%d: unsupported sector size %d.\n",
			     DEVICE_NAME, sdev->id, sector_size);
		}

	}

	scsi_release_request(sreq);
	scsi_free(buf, 512);
	return capacity;

}


/*
 *    bccfg_finsh() - Complete the initialization by registering the 
 *    the bccfg device as a disk and initializing it.
 */
static void
bccfg_finish()
{
	struct bccfg_dev *bccfg;
	int i;

	blk_dev[bccfg_major].queue = bccfg_find_queue;

	for (i = 0, bccfg = bccfg_devs; i < bccfg_template.dev_max; ++i, bccfg++) {
		if (!bccfg->capacity && bccfg->device) {

			DBG_PRINTK("bccfg_finish: Initializing %s%d\n",
				   DEVICE_NAME, i);

			if (bccfg_test_unit_ready(bccfg->device, 0)) {
			        bccfg->capacity = bccfg_read_capacity(bccfg->device, 0);

			}
		}
	}
	return;
}


#ifdef BCCFG_ALLOW_IO  /*   Don't allow reads or writes to the /dev/bccfg block
                        *   device by default, config commands must use the SG_IO
                        *   ioctl.
                        */ 

/*
 *    bccfg_init_command() - Call to initialize a SCSI Command for a 
 *    block device read or write operation.
 */
static int
bccfg_init_command(Scsi_Cmnd *SCpnt)
{
	int minor;
	int start_sector, sectors;
	/* struct hd_struct *ppnt; */
	struct bccfg_dev *bccfg;

	minor = DEVICE_NR(SCpnt->request.rq_dev);
     
	bccfg = &bccfg_devs[minor];

	start_sector = SCpnt->request.sector;
	sectors = SCpnt->request_bufflen >> 9;

	DBG_PRINTK("bccfg_init_command: Performing request, %s%d, start = %d, blocks = %d\n",
	           DEVICE_NAME, minor, start_sector, sectors);

	if (minor >= bccfg_template.dev_max ||
	    !bccfg->device ||
	    !bccfg->device->online ||
	    start_sector + SCpnt->request.nr_sectors > bccfg->capacity) {
		printk(KERN_WARNING "Failed bccfg request for device %s%d",
		       DEVICE_NAME, minor);
		return 0;
	}

	/*
	 * If we have a 1K hardware sectorsize, prevent access to single
	 * 512 byte sectors.  In theory we could handle this - in fact
	 * the scsi cdrom driver must be able to handle this because
	 * we typically use 1K blocksizes, and cdroms typically have
	 * 2K hardware sectorsizes.  Of course, things are simpler
	 * with the cdrom, since it is read-only.  For performance
	 * reasons, the filesystems should be able to handle this
	 * and not force the scsi disk driver to use bounce buffers
	 * for this.
	 */
	if (bccfg->device->sector_size == 1024) {
		if ((start_sector & 1) || (SCpnt->request.nr_sectors & 1)) {
			printk("bccfg_init_command: Bad block number requested");
			return 0;
		} else {
			start_sector = start_sector >> 1;
			sectors = sectors >> 1;
		}
	}
	if (bccfg->device->sector_size == 2048) {
		if ((start_sector & 3) || (SCpnt->request.nr_sectors & 3)) {
			printk("bccfg_init_command: Bad block number requested");
			return 0;
		} else {
			start_sector = start_sector >> 2;
			sectors = sectors >> 2;
		}
	}
	if (bccfg->device->sector_size == 4096) {
		if ((start_sector & 7) || (SCpnt->request.nr_sectors & 7)) {
			printk("bccfg_init_command: Bad block number requested");
			return 0;
		} else {
			start_sector = start_sector >> 3;
			sectors = sectors >> 3;
		}
	}

	switch (SCpnt->request.cmd) {

	case WRITE:
		SCpnt->cmnd[0] = WRITE_10;
		SCpnt->sc_data_direction = SCSI_DATA_WRITE;
		break;
	case READ:
		SCpnt->cmnd[0] = READ_10;
		SCpnt->sc_data_direction = SCSI_DATA_READ;
		break;
	default:
		panic("Unknown bccfg command %d\n", SCpnt->request.cmd);
	}
     
	DBG_PRINTK("bccfg_init_command: %s %d/%ld 512 byte blocks.\n",
		   (SCpnt->request.cmd == WRITE) ? "writing" : "reading",
		   sectors, SCpnt->request.nr_sectors);
     
	SCpnt->cmnd[1] = (SCpnt->device->scsi_level <= SCSI_2) ?
			 ((SCpnt->lun << 5) & 0xe0) : 0;
     
	SCpnt->cmnd[2] = (unsigned char) (start_sector >> 24) & 0xff;
	SCpnt->cmnd[3] = (unsigned char) (start_sector >> 16) & 0xff;
	SCpnt->cmnd[4] = (unsigned char) (start_sector >> 8) & 0xff;
	SCpnt->cmnd[5] = (unsigned char) start_sector & 0xff;
	SCpnt->cmnd[6] = SCpnt->cmnd[9] = 0;
	SCpnt->cmnd[7] = (unsigned char) (sectors >> 8) & 0xff;
	SCpnt->cmnd[8] = (unsigned char) sectors & 0xff;
     
	SCpnt->transfersize = bccfg->device->sector_size;
	SCpnt->underflow = sectors << 9;
     
	SCpnt->allowed = BCCFG_MAX_RETRIES;
	SCpnt->timeout_per_command = BCCFG_TIMEOUT;
     
	SCpnt->done = bccfg_rw_intr;

	return 1;
}



/*
 * bccfg_rw_intr is the interrupt routine for the device driver.
 * It will be notified on the end of a SCSI read / write, and
 * will take one of several actions based on success or failure.
 */

static void bccfg_rw_intr(Scsi_Cmnd * SCpnt)
{
	int result = SCpnt->result;
	int this_count = SCpnt->bufflen >> 9;
	int good_sectors = (result == 0 ? this_count : 0);
	int block_sectors = 1;
	long error_sector;

	DBG_PRINTK("bccfg_rw_intr; result %0x\n", result);

	/*
	   Handle MEDIUM ERRORs that indicate partial success.  Since this is a
	   relatively rare error condition, no care is taken to avoid
	   unnecessary additional work such as memcpy's that could be avoided.
	 */

	/* An error occurred */
	if (driver_byte(result) != 0 && 	/* An error occured */
	    (SCpnt->sense_buffer[0] & 0x7f) == 0x70) {	/* Sense data is valid */
		switch (SCpnt->sense_buffer[2]) {
		case MEDIUM_ERROR:
			if (!(SCpnt->sense_buffer[0] & 0x80))
				break;
			error_sector = (SCpnt->sense_buffer[3] << 24) |
			(SCpnt->sense_buffer[4] << 16) |
			(SCpnt->sense_buffer[5] << 8) |
			SCpnt->sense_buffer[6];
			if (SCpnt->request.bh != NULL)
				block_sectors = SCpnt->request.bh->b_size >> 9;
			switch (SCpnt->device->sector_size) {
			case 1024:
				error_sector <<= 1;
				if (block_sectors < 2)
					block_sectors = 2;
				break;
			case 2048:
				error_sector <<= 2;
				if (block_sectors < 4)
					block_sectors = 4;
				break;
			case 4096:
				error_sector <<=3;
				if (block_sectors < 8)
					block_sectors = 8;
				break;
			case 256:
				error_sector >>= 1;
				break;
			default:
				break;
			}
			error_sector &= ~(block_sectors - 1);
			good_sectors = error_sector - SCpnt->request.sector;
			if (good_sectors < 0 || good_sectors >= this_count)
				good_sectors = 0;
			break;

		case RECOVERED_ERROR:
			/*
			 * An error occured, but it recovered.  Inform the
			 * user, but make sure that it's not treated as a
			 * hard error.
			 */
			print_sense("bccfg", SCpnt);
			SCpnt->result = 0;
			SCpnt->sense_buffer[0] = 0x0;
			good_sectors = this_count;
			break;

		case ILLEGAL_REQUEST:
			if (SCpnt->device->ten == 1) {
				if (SCpnt->cmnd[0] == READ_10 ||
				    SCpnt->cmnd[0] == WRITE_10)
					SCpnt->device->ten = 0;
			}
			break;

		default:
			break;
		}
	}
	/*
	 * This calls the generic completion function, now that we know
	 * how many actual sectors finished, and how many sectors we need
	 * to say have failed.
	 */
	scsi_io_completion(SCpnt, good_sectors, block_sectors);
}
#else
static int
bccfg_init_command(Scsi_Cmnd *SCpnt)
{
  return 0;
}
#endif /* BCCFG_ALLOW_IO */



static int
bccfg_attach(struct scsi_device *sdev)
{
	unsigned int devnum;
	struct bccfg_dev *bccfg;
	int i;

	if (sdev->type != TYPE_PROCESSOR ||
	    memcmp(sdev->vendor, "RAIDCORE", 8) != 0)
		return 0;

	if (bccfg_template.nr_dev >= bccfg_template.dev_max) {
		sdev->attached--;
		return 1;
	}

	/* Find an empty minor device number. */
	for (bccfg = bccfg_devs, i = 0; i < bccfg_template.dev_max; i++, bccfg++)
		if (!bccfg->device)
			break;

	if (i >= bccfg_template.dev_max) {
		printk(KERN_WARNING "bccfg: Maximum device number excedded\n");
		sdev->attached--;
		return 1;
	}

	bccfg->device = sdev;
	bccfg_template.nr_dev++;

	printk("Attached scsi bccfg %s%d at scsi%d, channel %d, id %d, lun %d\n",
	       DEVICE_NAME, i,
	       sdev->host->host_no, sdev->channel, sdev->id, sdev->lun);

	return 0;
}


/*
 *    bccfg_detach() - called by the scsi layer remove a configuration device
 *    and release all of it resources.
 */
static void
bccfg_detach(struct scsi_device *sdev)
{
	struct bccfg_dev *bccfg;

	bccfg = bccfg_get_dev(sdev);

	if (bccfg) {
	  /* del_gendisk(&bccfg->gd); */
		bccfg->device = NULL;
		bccfg->capacity = 0;
		sdev->attached--;
		bccfg_template.dev_noticed--;
		bccfg_template.nr_dev--;
	}
}


/*	read routine for /proc/scsi/bccfg/version
 */
static int bccfg_proc_read(char *buf,	// read buffer (in kernel space)
			   char **start,// start of valid data in the buffer (see proc_file_read())
			   off_t offset,// Offset from the beginning of the imaginary file 
			   int buf_size,// bytes available
			   int *peof,	// set to 1 if we are at the end
			   void *data)	// proc_dirent->data
{
	int len;

	if (offset != 0)
		len = 0;
	else
		len = snprintf(buf, buf_size,
			       "V%s %s\n",
			       BCCFG_VERSION,
			       RC_BUILD_NUMBER);

	*peof = 1;
	*start = buf;
	return (len);
}


/*	write routine for /proc/scsi/bccfg/version
 */
static int bccfg_proc_write(struct file *file,	// our file control structure
			    const char *buf,    // the input buffer (user data address)
			    unsigned long count,// how many bytes we are being sent
			    void *data)         // proc_dirent->data
{
	return (-EINVAL);
}


static int __init
init_bccfg(void)
{

	extern char *bccfg_ident;
	int error;

	printk(KERN_NOTICE
           "Broadcom %s raid configuration driver version %s build_number %s built %s\n",
	       BCCFG_NAME,
	       BCCFG_VERSION,
	       RC_BUILD_NUMBER,
	       BCCFG_BUILD_DATE);
	printk(KERN_NOTICE "%s %s\n", BCCFG_NAME, bccfg_ident);

	bccfg_major = devfs_register_blkdev(0, DEVICE_NAME, &bccfg_bops);

	if (bccfg_major < 0) {
		printk(KERN_WARNING "bccfg: Couldn't get major device");
		bccfg_template.dev_noticed = 0;
		return -ENODEV;
	}

	printk(KERN_INFO "bccfg: Major device %d\n", bccfg_major);

	memset(bccfg_devs, 0, BCCFG_MAX_DEVS * sizeof(struct bccfg_dev));

	bccfg_template.module = THIS_MODULE;
	bccfg_template.major = bccfg_major;
	error = scsi_register_module(MODULE_SCSI_DEV, &bccfg_template);

	if (!error) {
		proc_bccfg = proc_mkdir("scsi/" BCCFG_NAME, NULL);
		if (proc_bccfg != NULL) {
			proc_bccfg_version = create_proc_entry("version", 0, proc_bccfg);
			if (proc_bccfg_version != NULL) {
				proc_bccfg_version->read_proc = bccfg_proc_read;
				proc_bccfg_version->write_proc = bccfg_proc_write;
				proc_bccfg_version->data = NULL;
			}
		}
	}

	return error;
}


static void __exit
exit_bccfg(void)
{

	if (proc_bccfg != NULL) {
		if (proc_bccfg_version != NULL) {
			remove_proc_entry("version", proc_bccfg);
			proc_bccfg_version = NULL;
		}
		remove_proc_entry("scsi/" BCCFG_NAME, NULL);
		proc_bccfg = NULL;
	}

	scsi_unregister_module(MODULE_SCSI_DEV, &bccfg_template);
	devfs_unregister_blkdev(bccfg_major, "bccfg");

}

module_init(init_bccfg);
module_exit(exit_bccfg);


