/*
 * bccfg.c - bcraid control processor scsi driver.
 *
 *   Provides configuration access to the bcraid control
 *   processor
 *
 * Copyright 2005 Broadcom Corporation. All rights reserved.
 * 
 * Redistribution and use of this Software in source and binary forms,
 * with or without modification, are permitted provided that the following
 * conditions are met:
 * 
 * 	1. Redistributions of source code must retain the above copyright
 *	   notice, this list of conditions and the following disclaimer. 
 *
 *	2. Redistributions in binary form must reproduce the above copyright
 * 	   notice, this list of conditions and the following disclaimer in
 *	   the documentation and/or other materials provided with the
 * 	   distribution.
 * 
 * 	3. This Software may be used only in connection with Broadcom
 * 	   storage integrated circuits.
 * 
 * THIS SOFTWARE IS PROVIDED BY BROADCOM CORPORATION AS IS AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL BROADCOM CORPORATION OR ITS LICENSORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be used, modified and/or distributed under the terms of
 * the GNU General Public License, version 2, as published by the Free
 * Software Foundation (the GPL), in which case the provisions of the GPL
 * apply INSTEAD OF those given above. A copy of the GPL is available at
 * http://www.gnu.org/copyleft/gpl.html, or by writing to the 
 *  	Free Software Foundation, Inc.
 *  	59 Temple Place - Suite 330
 *  	Boston, MA 02111-1307, USA.
 *
 *
 * This driver is derived from sd.c
 *      Copyright (C) 1992 Drew Eckhardt
 *      Copyright (C) 1993, 1994, 1995, 1999 Eric Youngdale
 *
 */

#include "version.h"
#include "build_number.h"


#include <linux/version.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/bio.h>
#include <linux/genhd.h>
#include <linux/hdreg.h>
#include <linux/errno.h>
#include <linux/idr.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/blkdev.h>
#include <linux/blkpg.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>

#include <scsi/scsi.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_driver.h>
#include <scsi/scsi_ioctl.h>
#include <scsi/scsi_request.h>
#include <scsi/scsi_eh.h>
#include <scsi/scsi_host.h>
#include <scsi/scsicam.h>
#if 0
#include <scsi/scsi_dbg.h>
#endif

#define BCCFG_VERSION           RC_VERSION_STR
#define BCCFG_BUILD_DATE		 __DATE__
#define BCCFG_BUILD_TIME		 __TIME__
#define BCCFG_NAME				 "bccfg"
#define BCCFG_MAX_DEVS           8
#define BCCFG_MAX_RETRIES        5
#define BCCFG_TIMEOUT            (30 * HZ)


MODULE_AUTHOR("Broadcom, Inc");
MODULE_DESCRIPTION("Configuration driver for Broadcom raid controllers");
MODULE_LICENSE("Dual BSD/GPL");

static int debug = 0;
#ifdef module_param_named
module_param_named(debug, debug, int, 0444);
#else
MODULE_PARM (debug, "i");
#endif
MODULE_PARM_DESC (debug, "debug print level");

#define DBG_PRINTK(fmt, ...) if (debug) printk(fmt, __VA_ARGS__);

/* support for /proc/scsi/bccfg/version
 */
static struct proc_dir_entry *proc_bccfg = NULL;
static struct proc_dir_entry *proc_bccfg_version = NULL;
static int bccfg_proc_read(char *page, char **start, off_t off, int count, int *peof, void *data);
static int bccfg_proc_write(struct file *file, const char *buf, unsigned long count, void *data);

static int bccfg_major = -1;  /* Major device number, dynamically assigned */

struct bccfg {
	struct scsi_driver *driver;	/* always &bccfg_template */
	struct scsi_device *device;
	atomic_t	       refcount;
	struct gendisk	   *disk;
	unsigned int	   openers;	    /* protected by BKL for now, yuck */
	sector_t	       capacity;	/* size in 512-byte sectors */
	u32		   index;
};

/* static struct idr bccfg_index_idr; */
static int indexes = 0;
static spinlock_t bccfg_index_lock = SPIN_LOCK_UNLOCKED;


/* This semaphore is used to mediate the 0->1 reference get in the
 * face of object destruction (i.e. we can't allow a get on an
 * object after last put) */
static DECLARE_MUTEX(bccfg_ref_sem);

static void bccfg_disk_release(struct bccfg *bccfg);

static int bccfg_revalidate_disk(struct gendisk *disk);
static void bccfg_rw_intr(struct scsi_cmnd * SCpnt);

static int bccfg_probe(struct device *);
static int bccfg_remove(struct device *);
static void bccfg_shutdown(struct device *dev);
static void bccfg_rescan(struct device *);
static int bccfg_init_command(struct scsi_cmnd *);
static void bccfg_read_capacity(struct bccfg *bccfgkp, char *diskname,
				unsigned char *buffer);

static struct scsi_driver bccfg_template = {
	.owner			= THIS_MODULE,
	.gendrv = {
		.name		= "bccfg",
		.probe		= bccfg_probe,
		.remove		= bccfg_remove,
		.shutdown	= bccfg_shutdown,
	},
	.rescan			= bccfg_rescan,
	.init_command	= bccfg_init_command,
};


static inline struct bccfg *bccfg(struct gendisk *disk)
{
	return container_of(disk->private_data, struct bccfg, driver);
}


static struct bccfg *bccfg_get(struct gendisk *disk)
{
	struct bccfg *bccfgp = NULL;

	down(&bccfg_ref_sem);
	if (disk->private_data == NULL)
		goto out;
	bccfgp = bccfg(disk);
	atomic_inc(&bccfgp->refcount);
	if (scsi_device_get(bccfgp->device))
		goto out_put;
	up(&bccfg_ref_sem);
	return bccfgp;

 out_put:
 	if (atomic_dec_and_test(&bccfgp->refcount))
		bccfg_disk_release(bccfgp);
	bccfgp = NULL;
 out:
	up(&bccfg_ref_sem);
	return bccfgp;
}


static void bccfg_put(struct bccfg *bccfgp)
{
	struct scsi_device *sdev = bccfgp->device;

	down(&bccfg_ref_sem);
	if (atomic_dec_and_test(&bccfgp->refcount))
		bccfg_disk_release(bccfgp);
	scsi_device_put(sdev);
	up(&bccfg_ref_sem);
}

/*
 * bccfg_get_index() - returns a index for the bccfg device
 * or BCCFG_MAX_DEVS if they are all taken.
 */
static int
bccfg_get_index(void)
{
        u32 i;
        u32 mask = 1;

        spin_lock(&bccfg_index_lock);

        for (i = 0, mask = 1; i < BCCFG_MAX_DEVS; i++, mask <<= 1) {
                if ((indexes & mask) == 0) {
                        indexes |= mask;
                        break;
                }    
        }
        spin_unlock(&bccfg_index_lock);
        return i;
}

/*
 * bccfg_put_index() - frees a previously reserved index
 */
static void
bccfg_put_index(u32 index)
{
     u32 mask = 1;

     if (index < BCCFG_MAX_DEVS) {
          mask <<= index;
          indexes &= ~mask;
     }
     else {
          DBG_PRINTK("bccfg_put_index: Index %d out of range.\n", index);
     }
     
}


/**
 *	bccfg_init_command - build a scsi (read or write) command from
 *	information in the request structure.
 *	@SCpnt: pointer to mid-level's per scsi command structure that
 *	contains request and into which the scsi command is written
 *
 *	Returns 1 if successful and 0 if error (or cannot be done now).
 **/
static int bccfg_init_command(struct scsi_cmnd * SCpnt)
{
	unsigned int this_count, timeout;
	struct request *rq = SCpnt->request;

	/* Only processes requests from SG_IO ioctl, return I/O error
	 * on reads or writes to the block device.
	 */
	if (!blk_pc_request(rq)) 
		return 0;

	if (sizeof(rq->cmd) > sizeof(SCpnt->cmnd))
		return 0;

	memcpy(SCpnt->cmnd, rq->cmd, sizeof(SCpnt->cmnd));
	if (rq_data_dir(rq) == WRITE)
		SCpnt->sc_data_direction = DMA_TO_DEVICE;
	else if (rq->data_len)
		SCpnt->sc_data_direction = DMA_FROM_DEVICE;
	else
		SCpnt->sc_data_direction = DMA_NONE;

	this_count = rq->data_len;
	if (rq->timeout)
		timeout = rq->timeout;
    else 
         timeout = BCCFG_TIMEOUT;
 
	SCpnt->transfersize = rq->data_len;
	SCpnt->allowed = BCCFG_MAX_RETRIES;
	SCpnt->timeout_per_command = timeout;

	/*
	 * This is the completion routine we use.  This is matched in terms
	 * of capability to this function.
	 */
	SCpnt->done = bccfg_rw_intr;

	/*
	 * This indicates that the command is ready from our end to be
	 * queued.
	 */
	return 1;
}


/**
 *	bccfg_open - open a scsi disk device
 *	@inode: only i_rdev member may be used
 *	@filp: only f_mode and f_flags may be used
 *
 *	Returns 0 if successful. Returns a negated errno value in case 
 *	of error.
 *
 *	Note: This can be called from a user context (e.g. fsck(1) )
 *	or from within the kernel (e.g. as a result of a mount(1) ).
 *	In the latter case @inode and @filp carry an abridged amount
 *	of information as noted above.
 **/
static int bccfg_open(struct inode *inode, struct file *filp)
{
	struct gendisk *disk = inode->i_bdev->bd_disk;
	struct bccfg *bccfgp;
	struct scsi_device *sdev;

	if (!(bccfgp = bccfg_get(disk)))
		return -ENXIO;

	DBG_PRINTK("bccfg_open: disk=%s\n", disk->disk_name);

	sdev = bccfgp->device;

	/*
	 * If the device is in error recovery, wait until it is done.
	 * If the device is offline, then disallow any access to it.
	 */
	if (!scsi_block_when_processing_errors(sdev)) {
		bccfg_put(bccfgp);
		return -ENXIO;	
        }

	return 0;
}

/**
 *	bccfg_release - invoked when the (last) close(2) is called on this
 *	scsi disk.
 *	@inode: only i_rdev member may be used
 *	@filp: only f_mode and f_flags may be used
 *
 *	Returns 0. 
 *
 *	Note: may block (uninterruptible) if error recovery is underway
 *	on this disk.
 **/
static int bccfg_release(struct inode *inode, struct file *filp)
{
	struct gendisk *disk = inode->i_bdev->bd_disk;
	struct bccfg *bccfgp = bccfg(disk);

	DBG_PRINTK("bccfg_release: disk=%s\n", disk->disk_name);

	bccfg_put(bccfgp);
	return 0;
}

static int bccfg_hdio_getgeo(struct block_device *bdev,
                              struct hd_geometry __user *loc)
{
	struct bccfg *bccfgp = bccfg(bdev->bd_disk);
	struct scsi_device *sdev = bccfgp->device;
	struct Scsi_Host *host = sdev->host;
	int diskinfo[4];

	/* default to most commonly used values */
	diskinfo[0] = 0x40;	/* 1 << 6 */
	diskinfo[1] = 0x20;	/* 1 << 5 */
	diskinfo[2] = bccfgp->capacity >> 11;
	
	/* override with calculated, extended default, or driver values */
	if (host->hostt->bios_param)
		host->hostt->bios_param(sdev, bdev, bccfgp->capacity, diskinfo);
	else
		scsicam_bios_param(bdev, bccfgp->capacity, diskinfo);

	if (put_user(diskinfo[0], &loc->heads))
		return -EFAULT;
	if (put_user(diskinfo[1], &loc->sectors))
		return -EFAULT;
	if (put_user(diskinfo[2], &loc->cylinders))
		return -EFAULT;
	if (put_user((unsigned)get_start_sect(bdev),
		     (unsigned long __user *)&loc->start))
		return -EFAULT;
	return 0;
}

/**
 *	bccfg_ioctl - process an ioctl
 *	@inode: only i_rdev/i_bdev members may be used
 *	@filp: only f_mode and f_flags may be used
 *	@cmd: ioctl command number
 *	@arg: this is third argument given to ioctl(2) system call.
 *	Often contains a pointer.
 *
 *	Returns 0 if successful (some ioctls return postive numbers on
 *	success as well). Returns a negated errno value in case of error.
 *
 *	Note: most ioctls are forward onto the block subsystem or further
 *	down in the scsi subsytem.
 **/
static int bccfg_ioctl(struct inode * inode, struct file * filp, 
		       unsigned int cmd, unsigned long arg)
{
	struct block_device *bdev = inode->i_bdev;
	struct gendisk *disk = bdev->bd_disk;
	struct scsi_device *sdev = bccfg(disk)->device;
	void __user *p = (void __user *)arg;
	int error;
 
	DBG_PRINTK("bccfg_ioctl: disk=%s, cmd=0x%x\n", disk->disk_name, cmd);

	/*
	 * If we are in the middle of error recovery, don't let anyone
	 * else try and use this device.  Also, if error recovery fails, it
	 * may try and take the device offline, in which case all further
	 * access to the device is prohibited.
	 */
	if (!scsi_block_when_processing_errors(sdev))
		return -ENODEV;

	if (cmd == HDIO_GETGEO) {
		if (!arg)
			return -EINVAL;
		return bccfg_hdio_getgeo(bdev, p);
	}

	/*
	 * Send SCSI addressing ioctls directly to mid level, send other
	 * ioctls to block level and then onto mid level if they can't be
	 * resolved.
	 */
	switch (cmd) {
		case SCSI_IOCTL_GET_IDLUN:
		case SCSI_IOCTL_GET_BUS_NUMBER:
			return scsi_ioctl(sdev, cmd, p);
		default:
             /* SUSE changed the APIs for a patch release. */
#if defined SCSI_CMD_IOCTL_FILP
			error = scsi_cmd_ioctl(filp, disk, cmd, arg);
#elif LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,7)
            error = scsi_cmd_ioctl(disk, cmd, arg);
#else
			error = scsi_cmd_ioctl(filp, disk, cmd, p);
#endif
			if (error != -ENOTTY)
				return error;
	}
	return scsi_ioctl(sdev, cmd, p);
}


static void bccfg_rescan(struct device *dev)
{
	struct bccfg *bccfgp = dev_get_drvdata(dev);
	bccfg_revalidate_disk(bccfgp->disk);
}

static struct block_device_operations bccfg_fops = {
	.owner			= THIS_MODULE,
	.open			= bccfg_open,
	.release	        = bccfg_release,
	.ioctl			= bccfg_ioctl,
	.revalidate_disk	        = bccfg_revalidate_disk,
};

/**
 *	bccfg_rw_intr - bottom half handler: called when the lower level
 *	driver has completed (successfully or otherwise) a scsi command.
 *	@SCpnt: mid-level's per command structure.
 *
 *	Note: potentially run from within an ISR. Must not block.
 **/
static void
bccfg_rw_intr(struct scsi_cmnd * SCpnt)
{
	int result = SCpnt->result;
	int this_count = SCpnt->bufflen;
	int good_bytes = (result == 0 ? this_count : 0);
	sector_t block_sectors = 1;
	sector_t error_sector;

	/*
	   Handle MEDIUM ERRORs that indicate partial success.  Since this is a
	   relatively rare error condition, no care is taken to avoid
	   unnecessary additional work such as memcpy's that could be avoided.
	 */

	/* An error occurred */
	if (driver_byte(result) != 0 && 	/* An error occurred */
	    (SCpnt->sense_buffer[0] & 0x7f) == 0x70) { /* Sense current */
		switch (SCpnt->sense_buffer[2]) {
		case MEDIUM_ERROR:
			if (!(SCpnt->sense_buffer[0] & 0x80))
				break;
			if (!blk_fs_request(SCpnt->request))
				break;
			error_sector = (SCpnt->sense_buffer[3] << 24) |
				       (SCpnt->sense_buffer[4] << 16) |
				       (SCpnt->sense_buffer[5] << 8) |
				       SCpnt->sense_buffer[6];
			if (SCpnt->request->bio != NULL)
				block_sectors = bio_sectors(SCpnt->request->bio);
			switch (SCpnt->device->sector_size) {
			case 1024:
				error_sector <<= 1;
				if (block_sectors < 2)
					block_sectors = 2;
				break;
			case 2048:
				error_sector <<= 2;
				if (block_sectors < 4)
					block_sectors = 4;
				break;
			case 4096:
				error_sector <<=3;
				if (block_sectors < 8)
					block_sectors = 8;
				break;
			case 256:
				error_sector >>= 1;
				break;
			default:
				break;
			}

			error_sector &= ~(block_sectors - 1);
			good_bytes = (error_sector - SCpnt->request->sector) << 9;
			if (good_bytes < 0 || good_bytes >= this_count)
				good_bytes = 0;
			break;

		case RECOVERED_ERROR: /* an error occurred, but it recovered */
		case NO_SENSE: /* LLDD got sense data */
			/*
			 * Inform the user, but make sure that it's not treated
			 * as a hard error.
			 */
			/* scsi_print_sense("sd", SCpnt); */
			SCpnt->result = 0;
			SCpnt->sense_buffer[0] = 0x0;
			good_bytes = this_count;
			break;

		case ILLEGAL_REQUEST:
			if (SCpnt->device->use_10_for_rw &&
			    (SCpnt->cmnd[0] == READ_10 ||
			     SCpnt->cmnd[0] == WRITE_10))
				SCpnt->device->use_10_for_rw = 0;
			if (SCpnt->device->use_10_for_ms &&
			    (SCpnt->cmnd[0] == MODE_SENSE_10 ||
			     SCpnt->cmnd[0] == MODE_SELECT_10))
				SCpnt->device->use_10_for_ms = 0;
			break;

		default:
			break;
		}
	}
	/*
	 * This calls the generic completion function, now that we know
	 * how many actual sectors finished, and how many sectors we need
	 * to say have failed.
	 */
	scsi_io_completion(SCpnt, good_bytes, block_sectors << 9);
}



/*
 * read disk capacity
 */
static void
bccfg_read_capacity(struct bccfg *bccfgp,
                    char *diskname,
                    unsigned char *buffer)
{
	unsigned char cmd[16];
	int the_result, retries;
	int sector_size = 0;
	int longrc = 0;
	struct scsi_device *sdev = bccfgp->device;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15)
	struct scsi_request *SRpnt;

	SRpnt = scsi_allocate_request(sdev, GFP_KERNEL);
	if (!SRpnt) {
		printk(KERN_WARNING
		       "bccfg_read_capacity: Request allocation failure.\n");
		return;
	}
#else
	struct scsi_sense_hdr sshdr;
#endif

repeat:
	retries = 3;
	do {
		if (longrc) {
			memset((void *) cmd, 0, 16);
			cmd[0] = SERVICE_ACTION_IN;
			cmd[1] = SAI_READ_CAPACITY_16;
			cmd[13] = 12;
			memset((void *) buffer, 0, 12);
		} else {
			cmd[0] = READ_CAPACITY;
			memset((void *) &cmd[1], 0, 9);
			memset((void *) buffer, 0, 8);
		}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15)
		SRpnt->sr_cmd_len = 0;
		memset(SRpnt->sr_sense_buffer, 0, SCSI_SENSE_BUFFERSIZE);
		SRpnt->sr_data_direction = DMA_FROM_DEVICE;

		scsi_wait_req(SRpnt, (void *) cmd, (void *) buffer,
			      longrc ? 12 : 8,
			      BCCFG_TIMEOUT, BCCFG_MAX_RETRIES);

		the_result = SRpnt->sr_result;
#else
		the_result = scsi_execute_req(sdev, cmd, DMA_FROM_DEVICE,
					      buffer, longrc ? 12 : 8, &sshdr,
					      BCCFG_TIMEOUT, BCCFG_MAX_RETRIES);
#endif

		retries--;

	} while (the_result && retries);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15)
	scsi_release_request(SRpnt);
#endif

	if (the_result && !longrc) {
		printk(KERN_NOTICE "%s : READ CAPACITY failed.\n"
		       "%s : status=%x, message=%02x, host=%d, driver=%02x \n",
		       diskname, diskname,
		       status_byte(the_result),
		       msg_byte(the_result),
		       host_byte(the_result),
		       driver_byte(the_result));

		if (driver_byte(the_result) & DRIVER_SENSE)
             /* scsi_print_req_sense("bccfg", SRpnt); */ ;
		else
			printk(KERN_NOTICE "%s : sense not available. \n", diskname);

		/* Either no media are present but the drive didn't tell us,
		   or they are present but the read capacity command fails */
		/* bccfgkp->media_present = 0; -- not always correct */
		bccfgp->capacity = 0x200000; /* 1 GB - random */

		return;
	} else if (the_result && longrc) {
		/* READ CAPACITY(16) has been failed */
		printk(KERN_NOTICE "%s : READ CAPACITY(16) failed.\n"
		       "%s : status=%x, message=%02x, host=%d, driver=%02x \n",
		       diskname, diskname,
		       status_byte(the_result),
		       msg_byte(the_result),
		       host_byte(the_result),
		       driver_byte(the_result));
		printk(KERN_NOTICE "%s : use 0xffffffff as device size\n",
		       diskname);

		bccfgp->capacity = 1 + (sector_t) 0xffffffff;		
		goto got_data;
	}	
	
	if (!longrc) {
		sector_size = (buffer[4] << 24) |
			(buffer[5] << 16) | (buffer[6] << 8) | buffer[7];
		if (buffer[0] == 0xff && buffer[1] == 0xff &&
		    buffer[2] == 0xff && buffer[3] == 0xff) {
			if (sizeof(bccfgp->capacity) > 4) {
				printk(KERN_NOTICE
				       "%s : very big device. try to use"
				       " READ CAPACITY(16).\n", diskname);
				longrc = 1;
				goto repeat;
			}
			printk(KERN_ERR "%s: too big for this kernel.  Use a "
			       "kernel compiled with support for large block "
			       "devices.\n", diskname);
			bccfgp->capacity = 0;
			goto got_data;
		}
		bccfgp->capacity = 1 + (((sector_t)buffer[0] << 24) |
				   (buffer[1] << 16) |
				   (buffer[2] << 8) |
			           buffer[3]);			
	} else {
		bccfgp->capacity = 1 + (((u64)buffer[0] << 56) |
				   ((u64)buffer[1] << 48) |
				   ((u64)buffer[2] << 40) |
				   ((u64)buffer[3] << 32) |
				   ((sector_t)buffer[4] << 24) |
				   ((sector_t)buffer[5] << 16) |
				   ((sector_t)buffer[6] << 8)  |
				   (sector_t)buffer[7]);

		sector_size = (buffer[8] << 24) |
			      (buffer[9] << 16) |
			      (buffer[10] << 8) |
			      buffer[11];
	}

got_data:
	if (sector_size == 0) {
		sector_size = 512;
		printk(KERN_NOTICE
		       "%s : sector size 0 reported, assuming 512.\n",
		       diskname);
	}

	if (sector_size != 512 &&
	    sector_size != 1024 &&
	    sector_size != 2048 &&
	    sector_size != 4096 &&
	    sector_size != 256) {
         printk(KERN_NOTICE "%s : unsupported sector size %d.\n",
		       diskname, sector_size);
		/*
		 * The user might want to re-format the drive with
		 * a supported sectorsize.  Once this happens, it
		 * would be relatively trivial to set the thing up.
		 * For this reason, we leave the thing in the table.
		 */
		bccfgp->capacity = 0;
		/*
		 * set a bogus sector size so the normal read/write
		 * logic in the block layer will eventually refuse any
		 * request on this device without tripping over power
		 * of two sector size assumptions
		 */
		sector_size = 512;
	}

	/* Rescale capacity to 512-byte units */
	if (sector_size == 4096)
		bccfgp->capacity <<= 3;
	else if (sector_size == 2048)
		bccfgp->capacity <<= 2;
	else if (sector_size == 1024)
		bccfgp->capacity <<= 1;
	else if (sector_size == 256)
		bccfgp->capacity >>= 1;

	bccfgp->device->sector_size = sector_size;
}



/*
 *  bccfg_revalidate_disk - called the first time a new disk is seen,
 *
 *  @disk: struct gendisk we care about
 */
static int bccfg_revalidate_disk(struct gendisk *disk)
{
	struct bccfg *bccfgp = bccfg(disk);
	struct scsi_device *sdev = bccfgp->device;
	unsigned char *buffer;

	DBG_PRINTK("bccfg_revalidate_disk: disk=%s\n", disk->disk_name);

	buffer = kmalloc(512, GFP_KERNEL | __GFP_DMA);
	if (!buffer) {
		printk(KERN_WARNING
		       "bccfg_revalidate_disk: Memory allocation failure.\n");
		return 0;
	}

	/* defaults, until the device tells us otherwise */
	sdev->sector_size = 512;
	bccfgp->capacity = 0;

	bccfg_read_capacity(bccfgp, disk->disk_name, buffer);

	set_capacity(disk, bccfgp->capacity);
	kfree(buffer);

	return 0;
}

/*
 *	bccfg_probe - called during driver initialization and whenever a
 *	new scsi device is attached to the system. It is called once
 *	for each scsi device (not just disks) present.
 *	@dev: pointer to device object
 *
 *	Returns 0 if successful (or not interested in this scsi device 
 *	(e.g. scanner)); 1 when there is an error.
 *
 *	Note: this function is invoked from the scsi mid-level.
 *	This function sets up the mapping between a given 
 *	<host,channel,id,lun> (found in bccfgp) and new device name 
 *	(e.g. /dev/bccfg0). More precisely it is the block device major 
 *	and minor number that is chosen here.
 *
 *	Assume bccfg_attach is not re-entrant (for time being)
 *	Also think about bccfg_attach() and bccfg_remove() running
 *      coincidentally.
 */
static int bccfg_probe(struct device *dev)
{
	struct scsi_device *sdev = to_scsi_device(dev);
	struct bccfg *bccfgp;
	struct gendisk *gd;
	int index = 0;
	int error;

	error = -ENODEV;
	/* TODO: Make sure it's a "RAIDCORE" before we take it. */
	if (sdev->type != TYPE_PROCESSOR)
		goto out;

	if (memcmp(sdev->vendor, "RAIDCORE", 8) != 0)
		goto out;

	printk(KERN_NOTICE "bccfg_attach: scsi device: <%d,%d,%d,%d>\n", 
	       sdev->host->host_no, sdev->channel, sdev->id, sdev->lun);

	error = -ENOMEM;
	bccfgp = kmalloc(sizeof(*bccfgp), GFP_KERNEL);
	if (!bccfgp)
		goto out;

	memset (bccfgp, 0, sizeof(*bccfgp));
	atomic_set(&bccfgp->refcount,1);

	gd = alloc_disk(1);
	if (!gd) {
		printk(KERN_WARNING "bccfg: Can't alloc disk.\n");
		goto out_free;
	}

    index = bccfg_get_index();

	if (index >= BCCFG_MAX_DEVS) {
		printk(KERN_WARNING
		       "bccfg: Maximum number of devices exceeded %x.\n", index);
		error = -EBUSY;
        goto out_put;
	}

	bccfgp->device = sdev;
	bccfgp->driver = &bccfg_template;
	bccfgp->disk = gd;
	bccfgp->index = index;
	bccfgp->openers = 0;

	gd->major = bccfg_major;
	gd->first_minor = index;
	gd->minors = 1;
	gd->fops = &bccfg_fops;

	sprintf(gd->disk_name, "bccfg%d", index);

	strcpy(gd->devfs_name, sdev->devfs_name);

	gd->private_data = &bccfgp->driver;

	bccfg_revalidate_disk(gd);

	gd->driverfs_dev = &sdev->sdev_gendev;
	gd->flags = GENHD_FL_DRIVERFS | GENHD_FL_SUPPRESS_PARTITION_INFO;
	gd->queue = bccfgp->device->request_queue;

	dev_set_drvdata(dev, bccfgp);
	add_disk(gd);

	printk(KERN_NOTICE "Attached Config Processor %s at scsi%d, channel %d, "
	       "id %d, lun %d\n",
	       gd->disk_name, sdev->host->host_no, sdev->channel,
	       sdev->id, sdev->lun);

	return 0;

out_put:
	put_disk(gd);
out_free:
	kfree(bccfgp);
out:
	return error;
}

/*	read routine for /proc/scsi/bccfg/version
 */
static int bccfg_proc_read(char *buf,	// read buffer (in kernel space)
			   char **start,// start of valid data in the buffer (see proc_file_read())
			   off_t offset,// Offset from the beginning of the imaginary file 
			   int buf_size,// bytes available
			   int *peof,	// set to 1 if we are at the end
			   void *data)	// proc_dirent->data
{
	int len;

	if (offset != 0)
		len = 0;
	else
		len = snprintf(buf, buf_size,
			       "V%s %s\n",
			       BCCFG_VERSION,
			       RC_BUILD_NUMBER);

	*peof = 1;
	*start = buf;
	return (len);
}


/*	write routine for /proc/scsi/bccfg/version
 */
static int bccfg_proc_write(struct file *file,	// our file control structure
			    const char *buf,    // the input buffer (user data address)
			    unsigned long count,// how many bytes we are being sent
			    void *data)         // proc_dirent->data
{
	return (-EINVAL);
}


/*
 *	bccfg_remove - called whenever a scsi disk (previously recognized by
 *	bccfg_probe) is detached from the system. It is called (potentially
 *	multiple times) during bccfg module unload.
 *	@bccfgp: pointer to mid level scsi device object
 *
 *	Note: this function is invoked from the scsi mid-level.
 *	This function potentially frees up a device name (e.g. /dev/bccfg0)
 *	that could be re-used by a subsequent bccfg_probe().
 */
static int bccfg_remove(struct device *dev)
{
	struct bccfg *bccfgp = dev_get_drvdata(dev);

    DBG_PRINTK("bccfg_remove() device %p.\n", dev);

	del_gendisk(bccfgp->disk);
	bccfg_shutdown(dev);
	down(&bccfg_ref_sem);
	if (atomic_dec_and_test(&bccfgp->refcount))
		bccfg_disk_release(bccfgp);
	up(&bccfg_ref_sem);

	return 0;
}

/**
 *	bccfg_disk_release - Called to free the bccfg structure
 *
 *	bccfg_ref_sem must be held entering this routine.  Because it is
 *	called on last put, you should always use the bccfg_get()
 *	bccfg_put() helpers which manipulate the semaphore directly
 *	and never directly manulipate the reference count
 **/
static void bccfg_disk_release(struct bccfg *bccfg)
{
	struct gendisk *disk = bccfg->disk;

    printk("bccfg_disk_release.\n");

    bccfg_put_index(bccfg->index);
    
	disk->private_data = NULL;

	put_disk(disk);

	kfree(bccfg);

    printk("bccfg_disk_release() - Exit.\n");
}

/*
 * Send a SYNCHRONIZE CACHE instruction down to the device through
 * the normal SCSI command structure.  Wait for the command to
 * complete.
 */
static void bccfg_shutdown(struct device *dev)
{
#if 0
	struct scsi_device *sdev = to_scsi_device(dev);
	struct bccfg *bccfgp = dev_get_drvdata(dev);
#endif

}	


/*
 *	init_bccfg - entry point for this driver (both when built in
 *                    or when a module).
 *
 *	Note: this function registers this driver with the scsi mid-level.
 */
static int __init init_bccfg(void)
{

	extern char *bccfg_ident;
	int error;

	printk(KERN_INFO
	       "Broadcom %s configuration driver version %s build_number %s built %s\n",
	       BCCFG_NAME, BCCFG_VERSION, RC_BUILD_NUMBER, BCCFG_BUILD_DATE);

	printk(KERN_INFO "%s %s\n", BCCFG_NAME, bccfg_ident);

	bccfg_major = register_blkdev(0, "bccfg");

	if (bccfg_major < 0)
		return -ENODEV;

	error = scsi_register_driver(&bccfg_template.gendrv);

	if (!error) {
		proc_bccfg = proc_mkdir("scsi/" BCCFG_NAME, NULL);
		if (proc_bccfg != NULL) {
			proc_bccfg_version = create_proc_entry("version", 0, proc_bccfg);
			if (proc_bccfg_version != NULL) {
				proc_bccfg_version->read_proc = bccfg_proc_read;
				proc_bccfg_version->write_proc = bccfg_proc_write;
				proc_bccfg_version->data = NULL;
			}
		}
	}

	return error;
}

/**
 *	exit_bccfg - exit point for this driver (when it is a module).
 *
 *	Note: this function unregisters this driver from the scsi mid-level.
 **/
static void __exit exit_bccfg(void)
{
	printk(KERN_INFO "exit_bccfg: exiting bccfg driver\n");

	if (proc_bccfg != NULL) {
		if (proc_bccfg_version != NULL) {
			remove_proc_entry("version", proc_bccfg);
			proc_bccfg_version = NULL;
		}
		remove_proc_entry("scsi/" BCCFG_NAME, NULL);
		proc_bccfg = NULL;
	}

	scsi_unregister_driver(&bccfg_template.gendrv);
	unregister_blkdev(bccfg_major, "bccfg");
}


module_init(init_bccfg);
module_exit(exit_bccfg);
