/*
 * Copyright 2005 Broadcom Corporation. All rights reserved.
 * 
 * Redistribution and use of this Software in source and binary forms,
 * with or without modification, are permitted provided that the following
 * conditions are met:
 * 
 * 	1. Redistributions of source code must retain the above copyright
 * 	   notice, this list of conditions and the following disclaimer. 
 * 
 * 	2. Redistributions in binary form must reproduce the above copyright
 * 	   notice, this list of conditions and the following disclaimer in
 * 	   the documentation and/or other materials provided with the
 * 	   distribution.
 * 
 * 	3. This Software may be used only in connection with Broadcom
 * 	   storage integrated circuits.
 * 
 * THIS SOFTWARE IS PROVIDED BY BROADCOM CORPORATION AS IS AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL BROADCOM CORPORATION OR ITS LICENSORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be used, modified and/or distributed under the terms of
 * the GNU General Public License, version 2, as published by the Free
 * Software Foundation (the GPL), in which case the provisions of the GPL
 * apply INSTEAD OF those given above. A copy of the GPL is available at
 * http://www.gnu.org/copyleft/gpl.html, or by writing to the 
 * 	Free Software Foundation, Inc.
 * 	59 Temple Place - Suite 330
 * 	Boston, MA 02111-1307, USA.
 */ 

/*
 *
 *   MODULE: bc_event.c
 *
 *	This module contains the code to handle events from the raid subsystem
 *
 *	Created By: Jon Fraser
 *
 */
#include "bc.h"
#include "linux/kmod.h"

void bc_cfg_change_detect(bc_uint32_t type, bc_uint32_t bus, int flags);
void bc_cfg_change_detect_tasklet( unsigned long arg);
void bc_cfg_change_response(struct bc_srb_s *srb);
void bc_event(bc_uint32_t type, bc_uint8_t bus, int update_mode);
void bc_send_inq(bc_uint32_t bus, bc_uint32_t target, bc_uint32_t lun, int update_mode);
void bc_inq_callback (bc_srb_t *srb);
void bc_notify_scsi_layer(int bus, int target, int lun, int present);
void bc_event_kthread(void *arg);

extern int bc_srb_seq_num;

typedef struct bc_event_thread_s {
	struct task_struct	thread;
	struct semaphore	ctl_sema;
	struct semaphore	sema;
	int			pid;
	int			running;
	int			stop;
	int			num_srb;
	bc_srb_queue_t		cfg_change_detect;
	bc_srb_queue_t		cfg_change_response;
	bc_uint32_t		targets[MAX_ARRAY];
} bc_event_thread_t;

bc_event_thread_t bc_event_thread;

int
bc_event_init(void)
{
	bc_event_thread_t *tp;

	tp = &bc_event_thread;

	memset(tp, 0, sizeof(bc_event_thread_t));

	/*
	 * setup tasklet to catch config change events;
	 */

	tp->cfg_change_detect.head = (bc_srb_t *)0;
	tp->cfg_change_detect.tail = (bc_srb_t *)0;
	spin_lock_init(&tp->cfg_change_detect.lock);

	tasklet_init(&tp->cfg_change_detect.tasklet, bc_cfg_change_detect_tasklet, (unsigned long)tp);

	/* 
	 * response queue
	 * Use a kernel thread so that the thread can sleep
	 */
	tp->cfg_change_response.head = (bc_srb_t *)0;
	tp->cfg_change_response.tail = (bc_srb_t *)0;
	spin_lock_init(&tp->cfg_change_response.lock);

	init_MUTEX_LOCKED(&tp->ctl_sema);
	init_MUTEX_LOCKED(&tp->sema);

        tp->pid = kernel_thread((int (*)(void *))bc_event_kthread, (void *)tp, 0);
        if(tp->pid < 0) {
                printk(KERN_ERR "bcraid: Unable to create event thread \n");
                return 0;
	}

	down(&tp->ctl_sema);
	return(1);
}

void
bc_event_shutdown(void)
{
	bc_event_thread_t *tp;

	tp = &bc_event_thread;
	tasklet_kill(&tp->cfg_change_detect.tasklet);

	/* stop event thread */
	tp->stop = 1;
	up(&tp->sema);
	down(&tp->ctl_sema);
}


void
bc_event(bc_uint32_t type, bc_uint8_t bus, int update_mode)
{
	bc_srb_t *srb;
	unsigned long irql;
	bc_event_thread_t *tp;

	tp = &bc_event_thread;

	switch (type) {
	    case BC_CTR_EVENT_CONFIG_CHANGE_DETECTED:
		bc_printk(BC_INFO, "bcraid: bc_event: config change detected on bus %d\n", bus);
		break;

	    case BC_CTR_EVENT_CONFIG_ARRAY_OFFLINE:
		bc_printk(BC_INFO, "bcraid: bc_event: array offline detected on bus %d\n", bus);
		break;

	    case BC_CTR_EVENT_CONFIG_DISK_OFFLINE:
		bc_printk(BC_INFO, "bcraid: bc_event: drive offline detected on bus %d\n", bus);
		break;
	
	    default:
		bc_printk(BC_WARN, "bcraid: bc_event: unknown event 0x%x detected on bus %d\n", type, bus);
		return;
	}

	if ((srb = kmalloc( sizeof(bc_srb_t), GFP_ATOMIC)) == 0) {
		bc_printk(BC_WARN,"bc_event: could not alloc mem for srb\n");
		bc_printk(BC_WARN,"bc_event: config change not processed\n");
		return;
	}

	srb->bus	= bus;
	srb->function	= type;
	srb->flags	= update_mode;
	srb->next	= (bc_srb_t *)0;

	spin_lock_irqsave(&tp->cfg_change_detect.lock, irql);

	if (tp->cfg_change_detect.head == (bc_srb_t *)0) {
		tp->cfg_change_detect.head = srb;
		tp->cfg_change_detect.tail = srb;
	} else {
		tp->cfg_change_detect.tail->next = srb;
		tp->cfg_change_detect.tail = srb;
	}
	spin_unlock_irqrestore(&tp->cfg_change_detect.lock, irql);

	tasklet_schedule(&tp->cfg_change_detect.tasklet);
}

void
bc_cfg_change_detect_tasklet( unsigned long arg)
{
	unsigned long	irql;
	bc_srb_t	*srb;
	bc_uint32_t	type, bus;
	int		update_mode;
	bc_event_thread_t *tp;

	tp = (bc_event_thread_t *)arg;

	spin_lock_irqsave(&tp->cfg_change_detect.lock, irql);

	while (tp->cfg_change_detect.head) {
		/*
		 * take an srb off the list
		 */
		srb = tp->cfg_change_detect.head;
		tp->cfg_change_detect.head = srb->next;
		srb->next = (bc_srb_t *)0;

		spin_unlock_irqrestore(&tp->cfg_change_detect.lock, irql);

		type	    = srb->function;
		bus 	    = srb->bus;
		update_mode = srb->flags;
		kfree(srb);

		bc_cfg_change_detect(type, bus, update_mode);
		spin_lock_irqsave(&tp->cfg_change_detect.lock, irql);
	}
	spin_unlock_irqrestore(&tp->cfg_change_detect.lock, irql);
}

void
bc_cfg_change_detect(bc_uint32_t type, bc_uint32_t bus, int update_mode)
{
	unsigned int i;

	bc_printk(BC_DEBUG,"bc_cfg_change_detect type 0x%x bus %d\n", type, bus);

	//
	// For a first pass, just test bus 0, targets 0-15, lun 0
	//
	for (i = 0; i <= 15; i++) {
		bc_send_inq(
			0,	/* bus    */
			i,	/* target */
			0,	/* lun    */
			update_mode);
	}
}

/*
 * Note: an inquiry could take 30 seconds or more when doing pass-through IO
 * and the scsi command has to time out for nonexistant devices.
 */
void
bc_send_inq(bc_uint32_t bus, bc_uint32_t target, bc_uint32_t lun, int update_mode)
{
	unsigned long	irql;
	bc_srb_t	*srb;
	int		size, sg_list_size;
	bc_softstate_t	*state;
	bc_inq_data_t	*inq_data;
	inq_cmd_t	*inq_cmd;
	bc_sg_list_t	*sg;

	state = &bc_state;

	if (bus > 0  || target > 15 || lun > 0) {
		bc_printk(BC_WARN, "bc_send_inq: invalid B/T/L %d/%d/%d\n",
			bus,
			target,
			lun );
		return;
	}

	sg_list_size =  sizeof (bc_sg_list_t) + (BC_SG_MAX_ELEMENTS-1) * sizeof(bc_sg_elem_t);
	size = sizeof(bc_srb_t) + sg_list_size + state->memsize_per_srb;

	if ((srb = kmalloc( size, GFP_ATOMIC)) == 0) {
		bc_printk(BC_WARN,"bc_cfg_change_detect: could not alloc %d bytes for srb\n", size);
		// FIXME
		return;
	}

	memset(srb, 0, size);

	size = sizeof(inq_cmd_t) + sizeof(bc_inq_data_t); 

	if ((inq_cmd = kmalloc(size, GFP_ATOMIC)) == 0) {	
		bc_printk(BC_WARN,"bc_cfg_change_detect: could not alloc mem for inquiry command and data\n");
		// FIXME
		kfree(srb);	
		return;
	}

	memset(inq_cmd, size, 0);
	inq_data = (bc_inq_data_t *) ((char *)inq_cmd + sizeof (inq_cmd_t));

	srb->function     = BC_SRB_EXECUTE_SCSI;
	srb->status       = BC_SRB_STATUS_SUCCESS;
	srb->bus          = bus;
	srb->target       = target;
	srb->lun          = lun;
	srb->flags        = BC_SRB_FLAGS_UNSPECIFIED_DIRECTION;
	srb->data_len     = sizeof(bc_inq_data_t);
	srb->cdb          = inq_cmd;
	srb->cdb_len      = sizeof(inq_cmd_t);
	srb->scsi_context = NULL;
	srb->callback	  = bc_inq_callback;  // FIXME

	srb->sg_list	  = (bc_sg_list_t *)&srb->private32[0];
	srb->dev_private  = (char *)srb->sg_list + sg_list_size;
	srb->next	  = (bc_srb_t *)0;
	srb->seq_num 	  = bc_srb_seq_num;
	bc_srb_seq_num++;

	if (update_mode == BC_SRB_LOCAL_UPDATE_ONLY)
		srb->flags |= BC_SRB_LOCAL_UPDATE_ONLY;

	/*
	 * Build the inquiry command.
	 * The command has already been memset to zero.
	 */

	inq_cmd->opcode = SS_INQUIR;
	inq_cmd->len = sizeof(bc_inq_data_t);

	/*
	 * build the sg list
	 */
	sg = srb->sg_list;
	sg->sg_num_elem = 1;
	sg->sg_mem_type = BC_MEM_VADDR;

	sg->sg_elem[0].size = sizeof(bc_inq_data_t);
	sg->sg_elem[0].v_addr = inq_data;


	bc_printk(BC_DEBUG, "bc_send_inq: seq_num: %d B/T/L %d/%d/%d\n", 
			srb->seq_num, bus, target, lun);

	/*
	 *   Add the srb to the tail of the queue, schedule the q tasklet, and return.
	 *   Sending srb's through the OSIC can take a long time. We don't want other
	 *   CPU's spinning on the OSIC lock and just wasting cycles.  So we queue the srb's
	 *   and just have one tasklet entering the OSIC.
	 */

	state->stats.scb_total++;
	state->stats.target_total[target]++;

	atomic_inc(&state->stats.scb_pending);
	atomic_inc(&state->stats.target_pending[target]);

	spin_lock_irqsave(&state->srb_q.lock, irql);

	if (state->srb_q.head == (bc_srb_t *)0) {
		state->srb_q.head = srb;
		state->srb_q.tail = srb;
	} else {
		state->srb_q.tail->next = srb;
		state->srb_q.tail = srb;
	}
	spin_unlock_irqrestore(&state->srb_q.lock, irql);

	tasklet_schedule(&state->srb_q.tasklet);
}

/*
 * Just queue the inquiry response to the bc_event_thread kernel thread
 */
void
bc_inq_callback (bc_srb_t *srb)
{
	unsigned long irql;
	int 		do_wakeup;
	bc_event_thread_t *tp;

	tp = &bc_event_thread;


	bc_printk(BC_DEBUG, "bc_inq_callback: target %d, \n", srb->target);

	srb->next = (bc_srb_t *)0;
	spin_lock_irqsave(&tp->cfg_change_response.lock, irql);

	if (tp->cfg_change_response.head == (bc_srb_t *)0) {
		tp->cfg_change_response.head = srb;
		tp->cfg_change_response.tail = srb;
	} else {
		tp->cfg_change_response.tail->next = srb;
		tp->cfg_change_response.tail = srb;
	}

	if (bc_event_thread.num_srb == 0)
		do_wakeup = 1;
	else
		do_wakeup = 0;
	bc_event_thread.num_srb++;

	spin_unlock_irqrestore(&tp->cfg_change_response.lock, irql);

	if (do_wakeup)
		up(&bc_event_thread.sema);

}

void
bc_event_kthread(void *bc_threadp)
{
	bc_event_thread_t	*tp;
	bc_srb_t 		*srb;
	bc_softstate_t		*state;
	unsigned long		irql;

	state = &bc_state;
	tp    = (bc_event_thread_t*)bc_threadp;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        daemonize("bcraid/event_thread");
#else
	daemonize();
#endif

        set_user_nice(current, -10);
        sigfillset(&current->blocked);


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
        sprintf(current->comm, "bcraid/event_thread");
#endif

	up(&tp->ctl_sema);

	bc_printk(BC_INFO, "bcraid: bc_event_kthread  thread started\n");

	while (1) {
	    spin_lock_irqsave(&tp->cfg_change_response.lock, irql);

	    while (tp->cfg_change_response.head) {
		    /*
		     * take an srb off the list
		     */
		    srb = tp->cfg_change_response.head;
		    tp->cfg_change_response.head = srb->next;
		    srb->next = (bc_srb_t *)0;
		    tp->num_srb--;

		    spin_unlock_irqrestore(&tp->cfg_change_response.lock, irql);
		    bc_cfg_change_response(srb);
		    spin_lock_irqsave(&tp->cfg_change_response.lock, irql);
	    }

	    spin_unlock_irqrestore(&tp->cfg_change_response.lock, irql);

	    bc_printk(BC_DEBUG,"bc_event_thread: sleeping\n");
	    tp->running = 0;

	    /* down() is counted in load average, down_interruptible() is not */
	    if (down_interruptible(&tp->sema))
		bc_printk(BC_DEBUG, "bc_event_thread: interrupted\n");

	    tp->running = 1;
	    bc_printk(BC_DEBUG,"bc_event_kthread: awake\n");

	    if (tp->stop == 1)
		break;
	}

	bc_printk(BC_INFO, "bcraid: bc_event_kthread stopped\n");
	up(&tp->ctl_sema);
}

void
bc_cfg_change_response(struct bc_srb_s *srb)
{
	bc_uint32_t  bus, target, lun, status;
	bc_inq_data_t	*inq_data;
	bc_sg_list_t	*sg;
	char		*msg;
	int		local_update;
	int		present, was_present;
	bc_softstate_t	*state;
	bc_event_thread_t *tp;

	tp = &bc_event_thread;
	state = &bc_state;

	status	= srb->status;
	bus	= srb->bus;
	target	= srb->target;
	lun	= srb->lun;
	sg 	= srb->sg_list;
	inq_data = (bc_inq_data_t *)sg->sg_elem[0].v_addr;

	if (srb->flags & BC_SRB_LOCAL_UPDATE_ONLY ) 
		local_update = 1;
	else
		local_update = 0;

	msg = "unknown status";
	present = 0;

	was_present =  tp->targets[target] & (1 << lun) ? 1 : 0;

	switch (status) {
	    case BC_SRB_STATUS_SUCCESS:
		present = 1;
		msg = "device present";
		break;

	
	    case BC_SRB_STATUS_NO_DEVICE:
		msg = "no device";
		break;

	    case BC_SRB_STATUS_INVALID_LUN:
		msg = "invalid lun";
		break;

	    case BC_SRB_STATUS_INVALID_TARGET_ID:
		msg = "invalid target";
		break;

	    case BC_SRB_STATUS_INVALID_PATH_ID: 
		msg = "invalid path";
		break;
	    
	    default:
	        break;
	}

	bc_printk(BC_DEBUG, "bc_cfg_change_response B/T/L %d/%d/%d  %s\n",
	    	bus, target, lun, msg);

	/* device present status changed */
	if (was_present != present)  {
		if (present)
		    tp->targets[target] |= (1 << lun);
		else
		    tp->targets[target] &= ~(1 << lun);

		if (local_update == 0)
			bc_notify_scsi_layer(bus, target, lun, present);
	}

	kfree(srb->cdb);	/* free inquiry command and response data */
	kfree(srb);
}

static unsigned char exec_buf[128];
void 
bc_notify_scsi_layer(int bus, int target, int lun, int present)
{
	bc_softstate_t *state;

//#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#ifdef not_working_right
	struct scsi_device *sdev;
	state = &bc_state;

	bc_printk(BC_INFO,"bcraid:  B/T/L %d/%d/%d was %s\n",
			bus, target, lun, present ? "attached" : "detached");

	if (present) {
		__scsi_add_device(state->host_ptr,  bus, target, lun, NULL);
		scsi_host_put(state->host_ptr);
		module_put(state->host_ptr->hostt->module);

	} else {
	        sdev = scsi_device_lookup(state->host_ptr, bus, target, lun);
		if (sdev) {
			scsi_remove_device(sdev);
			scsi_device_put(sdev);
		}
	}
#else
	char *argv[] = { "/bin/sh", "-c", exec_buf, 0};
	
	char *envp[3];
	int host, status;

	state = &bc_state;
	host = state->host_ptr->host_no;

	memset(exec_buf, 0, sizeof(exec_buf));

	bc_printk(BC_INFO,"bcraid:  B/T/L %d/%d/%d was %s\n",
			bus, target, lun, present ? "attached" : "detached");

	if (present) {
		snprintf(exec_buf,sizeof(exec_buf),
			"echo scsi add-single-device %d %d %d %d > /proc/scsi/scsi",
			host, bus, target, lun);
	} else {
		snprintf(exec_buf,sizeof(exec_buf),
			"echo scsi remove-single-device %d %d %d %d > /proc/scsi/scsi",
			host, bus, target, lun);
	}


	/* minimal command environment */
	envp[0] = "HOME=/";
	envp[1] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin";
	envp[2] = (char *)0;

	bc_printk(BC_WARN,"bcraid: %s %s %s\n", argv[0], argv[1], argv[2]);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	status = call_usermodehelper(argv[0], argv, envp, 1);
#else
	status = call_usermodehelper(argv[0], argv, envp);
#endif
	bc_printk(BC_DEBUG, "bcraid: call_usermodehelper returned %d (0x%x)\n", -status, -status);
#endif
}

