
/*
 * Copyright 2003-2004 Broadcom Corporation. All rights reserved.
 * 
 * Redistribution and use of this Software in source and binary forms,
 * with or without modification, are permitted provided that the following
 * conditions are met:
 * 
 * 	1. Redistributions of source code must retain the above copyright
 * 	   notice, this list of conditions and the following disclaimer. 
 * 
 * 	2. Redistributions in binary form must reproduce the above copyright
 * 	   notice, this list of conditions and the following disclaimer in
 * 	   the documentation and/or other materials provided with the
 * 	   distribution.
 * 
 * 	3. This Software may be used only in connection with Broadcom
 * 	   storage integrated circuits.
 * 
 * THIS SOFTWARE IS PROVIDED BY BROADCOM CORPORATION AS IS AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL BROADCOM CORPORATION OR ITS LICENSORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be used, modified and/or distributed under the terms of
 * the GNU General Public License, version 2, as published by the Free
 * Software Foundation (the GPL), in which case the provisions of the GPL
 * apply INSTEAD OF those given above. A copy of the GPL is available at
 * http://www.gnu.org/copyleft/gpl.html, or by writing to the 
 * 	Free Software Foundation, Inc.
 * 	59 Temple Place - Suite 330
 * 	Boston, MA 02111-1307, USA.
 */ 

/*
 * This is i386 arch dependent
 */

bc_uint64_t bc_fsave_data[256/sizeof (bc_uint64_t)];	/* should only need 108 bytes */
bc_uint32_t bc_fsave_cr0;				/* saved along with fsave */
bc_uint32_t bc_fsave_saved;				/* we've saved the fp state */
bc_uint32_t bc_mem_movntq;				/* can use prefetchnta/movntq/sfence opcodes */

/*
 * get processor id/flags info
 */
static bc_uint32_t __inline__
bc_cpuid(bc_uint32_t func)
{
    bc_uint32_t eax, edx;

    __asm__ __volatile__ (
	"cpuid\n\t"		/* in: eax  out: eax/ebx/ecx/edx */
	: "=d"(edx),		/* we want register edx */
	  "=a"(eax)		/* tell compiler eax gets clobbered */
	: "1"(func)		/* function code in register eax */
	: "ebx", "ecx"		/* also clobber ebx, ecx */
	);
    return (edx);
}

/*
 * get processor vendor string
 */
static void __inline__
bc_cpuid_vendor(char *v)
{
    __asm__ __volatile__ (
	"movl	$0,%%eax\n\t"	/* function code 0, get vendor string */
	"cpuid\n\t"		/* zaps eax/ebx/ecx/edx */
	"movl	%0,%%eax\n\t"
	"movl	%%ebx,0(%%eax)\n\t"	/* vendor comes back in ebx */
	"movl	%%edx,4(%%eax)\n\t" /* ...edx */
	"movl	%%ecx,8(%%eax)\n\t" /* ...and ecx */
	: 
	: "m"(v)
	: "eax", "ebx", "ecx", "edx"
	);
}

/*
 * standard feature flags
 * cpuid function 1, returned in edx
 */
#define BC_HAS_MMX		0x00800000
#define BC_HAS_FXSR		0x01000000 /* fast floating save/restore (xmm regs) */
#define BC_HAS_SSE		0x02000000
#define BC_HAS_SSE2		0x04000000

/*
 * extended feature flags
 * cpuid function 0x80000001, returned in edx
 */
#define BC_HAS_3DNOW		0x80000000 /* theoretically, not AMD specific */

/*
 * AMD specific extended feature flags
 * cpuid function 0x80000001, returned in edx
 * only valid if processor vendor string is "AuthenticAMD"
 */
#define BC_HAS_AMD_MMX_EXT	0x00400000 /* implied by HAS_SSE */
#define BC_HAS_AMD_3DNOW_EXT	0x40000000


/*
 * bc_memcpy_init
 * 
 * Determine which memcpy code will actually be used
 */
static void __inline__
bc_memcpy_init(void)
{
    char *vendor_amd = "AuthenticAMD";
    char vendor[12];
    bc_uint32_t i;

    bc_mem_movntq = 0;

    if (bc_cpuid(1) & BC_HAS_SSE){
        bc_mem_movntq = 1;
	return;
    }

    bc_cpuid_vendor(vendor);
    for (i = 0; i < 12; ++i){
	if (vendor[i] != vendor_amd[i]) {
	    return;
	}
    }
    if (bc_cpuid(0x80000001) & BC_HAS_AMD_MMX_EXT){
        bc_mem_movntq = 1;
    }
}



/*
 * bc_memcpy_enter
 * 
 * Save what state is needed before using bc_memcpy
 * In this case, save the FP registers
 */
static void __inline__
bc_memcpy_enter(void)
{
    if (bc_fsave_saved == 0) {
	__asm__ __volatile__ (
	    "movl  %%cr0, %%eax\n\t"
	    "movl %%eax, %0\n\t"
	    "andl  $0xfffffff3, %%eax\n\t"			/* clear EM, TS (trap if MMX/FP instr) */
	    "movl  %%eax, %%cr0\n\t"
	    "fsave %1\n"
	    : "=m"(bc_fsave_cr0), "=m"(bc_fsave_data)
	    );
    }

    bc_fsave_saved = 1;
}

/*
 * bc_memcpy_exit
 *
 * Restore state after bc_memcpy
 * In this case, restore FP registers
 */
static void __inline__
bc_memcpy_exit(void)
{
    __asm__ __volatile__ (
	"frstor %0\n\t"
	"movl %1,%%eax\n\t"
	"movl %%eax,%%cr0\n"
	: "=m"(bc_fsave_data)
	: "m"(bc_fsave_cr0) );

    bc_fsave_saved = 0;
}

#define BC_ALIGNED_4(x)  (((bc_uint32_t)(x) & 0x3) == 0)
#define BC_ALIGNED_8(x)  (((bc_uint32_t)(x) & 0x7) == 0)

/*
 * fast memcpy using SSE/AMD extensions to mmx: prefetchnta/movntq/sfence
 *
 * We preload the cache, and load some more in the loop ahead of when we will need it using
 * prefetchnta, and we store the data later with movntq to avoid cache pollution.
 *
 * This seems to be faster than using movl to load a word from each cache line, and
 * generally faster than rep movs on data that is not already in the cache.
 *
 * NB: this routine expects to be called with 8-byte aligned buffers, and a minimum
 * of 256 bytes to move.  It will only move multiples of 64 bytes, and will update
 * the arg list so the caller can move the rest.
 */
static void __inline__
bc_memcpy_movntq(void **to, void **from, bc_uint32_t *n)
{
    __asm__ __volatile__ (
	"movl	%0,%%edi\n\t"			/* dest -- 8-byte aligned */
	"movl	%1,%%esi\n\t"			/* source -- 8-byte aligned */
	"movl	%2,%%ecx\n\t"			/* count -- minimum of 320 bytes */
	//"emms\n\t"				/* empty mmx state, should not be necessary */
	
	"movl	%%ecx,%%eax\n\n"		/* save count */
	"shrl	$6,%%ecx\n\t"			/* count by 64-byte blocks (we know we have at least 5 blocks (320 bytes) to move) */
	"subl	$4,%%ecx\n\t"			/* we prefetch 4 blocks (256 bytes) ahead */
	"movl	$4,%%edx\n\t" 			/* so we have 4 blocks left over at the end */

#if 0
#define FETCH(a) "movl	" a "(%%esi),%%ebx\n\t"		/* force the fetch... processor won't ignore this */
#else 
#define FETCH(a) "prefetchnta " a "(%%esi)\n\t"		/* software prefetch, load 1 cache line */
#endif
	/* prefetch and move:
	 * prefetch 256 bytes, then prefetch the 64-byte block 256 bytes ahead
	 * and copy the current 64-byte block
	 * esi -- source
	 * edi -- dest
	 * eax -- number of bytes left to move when we are done with the block moves
	 * ebx -- used to "prefetch" a cache line if we don't use the prefetchnta instr
	 * ecx -- move loop count (# 64-byte blocks to move)
	 * edx -- # blocks to move after we stop prefetching
	 */
	FETCH("0")				/* prefetch for the initial 4 iterations */
	"andl	$63,%%eax\n\t"			/* update final "n" here since it will be free */
	FETCH("64")
	FETCH("128")
	FETCH("192")
	".p2align 3\n\t"
	"3:\n\t"				/* move/prefetch loop */
	FETCH("256")				/* prefetch another 64-byte block, 256 bytes ahead of where we are processing */
#undef FETCH
	"31:\n\t"				/* move-only loop */
	"movq	0(%%esi),%%mm0\n\t"
	"movq	8(%%esi),%%mm1\n\t"
	"movq	16(%%esi),%%mm2\n\t"
	"movq	24(%%esi),%%mm3\n\t"
	"movq	32(%%esi),%%mm4\n\t"
	"movq	40(%%esi),%%mm5\n\t"
	"movq	48(%%esi),%%mm6\n\t"
	"movq	56(%%esi),%%mm7\n\t"
	"addl	$64,%%esi\n\t"			/* bump source address */
	"movntq	%%mm0,0(%%edi)\n\t"
	"movntq	%%mm1,8(%%edi)\n\t"
	"movntq	%%mm2,16(%%edi)\n\t"
	"movntq	%%mm3,24(%%edi)\n\t"
	"movntq	%%mm4,32(%%edi)\n\t"
	"movntq	%%mm5,40(%%edi)\n\t"
	"movntq	%%mm6,48(%%edi)\n\t"
	"movntq	%%mm7,56(%%edi)\n\t"
	"addl	$64,%%edi\n\t"			/* bump dest address */
	
	"decl	%%ecx\n\t"
	"jg 	3b\n\t"				/* still more to move with prefetch, so do it */
	"decl	%%edx\n\t"
	"jge	31b\n\t"			/* still more to move without doing any more prefetch */
	
	"sfence\n\t"				/* make sure all movntq data is stable */
	"movl	%%eax,%2\n\t"			/* update "n" */
	"movl	%%edi,%0\n\t"			/* update "to" */
	"movl	%%esi,%1\n\t"			/* update "from" */
	//"emms\n\t"				/* empty mmx state, should not be necessary */
						/* got 0-63 bytes left at this point... fall through to byte move */
	: "=m"(*to), "=m"(*from), "=m"(*n)
	:
	: "memory", "edi", "esi", "eax", "ebx", "ecx", "edx");
}

/*
 * move by words (4 bytes), and then by bytes
 */
static void __inline__
bc_memcpy_movs(void *to, void *from, bc_uint32_t n)
{
    __asm__ __volatile__ (
	// rep takes count in ecx
	// movs takes source in esi, dest in edi, inc/dec flag in DF
	// do 4 bytes at a time
	"movl	%0,%%edi\n\t"
	"movl	%1,%%esi\n\t"
	"movl	%2,%%ecx\n\t"
	"shrl	$2,%%ecx\n\t"
	"cld\n\t"		// clear DF -- increasing copy
	"rep\n\t"
	"movsl\n\t"		// copy all the 4-byte words
	"movl	%2,%%ecx\n\t"	// copy the rest
	"andl	$3,%%ecx\n\t"
	"rep\n\t"
	"movsb\n\t"		// copy a byte at a time
	:
	: "m"(to), "m"(from), "m"(n)
	: "memory", "edi", "esi", "ecx");
}

/*
 * move by bytes for unaligned data
 */
static void __inline__
bc_memcpy_movsb(void *to, void *from, bc_uint32_t n)
{
    __asm__ __volatile__ (
	// rep takes count in ecx
	// movs takes source in esi, dest in edi, inc/dec flag in DF
	// do a byte at a time
	"movl	%0,%%edi\n\t"
	"movl	%1,%%esi\n\t"
	"movl	%2,%%ecx\n\t"
	"cld\n\t"		// clear DF -- increasing copy
	"rep\n\t"
	"movsb\n\t"		// copy a byte at a time
	:
	: "m"(to), "m"(from), "m"(n)
	: "memory", "edi", "esi", "ecx");
}	


/*
 * fast memcpy: move as much as we can with block moves
 * then move the rest by words or bytes depending on the alignment
 */
void  __inline__
bc_memcpy(void *to, void *from, bc_uint32_t n)
{
    /* bc_memcpy_movntq() will fail if n < 320 */
    if (bc_mem_movntq && BC_ALIGNED_8(to) && BC_ALIGNED_8(from) && n >= 320) {
	bc_memcpy_enter();
	bc_memcpy_movntq(&to, &from, &n);	/* move multiples of 64-byte blocks and update the args */
    }

    if (n == 0)
	return;

    if (BC_ALIGNED_4(to) && BC_ALIGNED_4(from))
	bc_memcpy_movs(to, from, n);
    else
	bc_memcpy_movsb(to, from, n);
}
